/********************************************************************
 * Copyright(c) 2000 Broadcom Corporation, all rights reserved.
 * Contains proprietary and confidential information.
 *
 * This source file is the property of Broadcom Corporation, and
 * may not be copied or distributed in any isomorphic form without 
 * the prior written consent of Broadcom Corporation. 
 *
 *
 *  Name: T 3 _ P C I . C
 *
 *  Description: This file contains PCI utilitily routines.
 *
 *  AUTHOR: Kevin Tran
 *
 *  HISTORY:    Kevin Tran        06/01/2000   Created
 *
 *******************************************************************/

/****************************************************************************/
/*                                                                          */
/* Module: PCILIB.C                                                         */
/*                                                                          */
/* Purpose: Define a C interface to the PCI BIOS                            */
/*                                                                          */
/* Functions Defined:                                                       */
/*                                                                          */
/*   PCI_BIOS_PRESENT                                                       */
/*   FIND_PCI_DEVICE                                                        */
/*   FIND_PCI_CLASS_CODE                                                    */
/*   READ_CONFIGURATION_BYTE                                                */
/*   READ_CONFIGURATION_WORD                                                */
/*   READ_CONFIGURATION_DWORD                                               */
/*   WRITE_CONFIGURATION_BYTE                                               */
/*   WRITE_CONFIGURATION_WORD                                               */
/*   WRITE_CONFIGURATION_DWORD                                              */
/*   OUTPD                                                                  */
/*   INPD                                                                   */
/*   INSB                                                                   */
/*   INSW                                                                   */
/*   INSD                                                                   */
/*   OUTSB                                                                  */
/*   OUTSW                                                                  */
/*   OUTSD                                                                  */
/*                                                                          */
/* Local Functions                                                          */
/*                                                                          */
/*   READ_CONFIGURATION_AREA                                                */
/*   WRITE_CONFIGURATION_AREA                                               */
/*   PCI_GENINTERRUPT                                                       */
/*                                                                          */
/****************************************************************************/

/****************************************************************************/
/*                                                                          */
/* Include Files                                                            */
/*                                                                          */
/****************************************************************************/

#include <i86.h>
#include <dos.h>
#include <conio.h>


#include <stddef.h>
#include <string.h>


#ifdef _B57UTIL
    #include "generic.h"
    #define MAX_DEVICES 16
#else
    #include "general.h"
#endif
#include "pci.h"
#include "t_io.h"

/****************************************************************************/
/*                                                                          */
/* Register struct for simulating real-mode interupt from protected mode    */
/*                                                                          */
/****************************************************************************/

/*
 * Global to track devices
 */
typedef struct bios_reg {
   u32 edi;
   u32 esi;
   u32 ebp;
   u32 reserved1;
   u32 ebx;
   u32 edx;
   u32 ecx;
   u32 eax;
   u16 flags;
   u16 es, ds, fs, gs, ip, cs, sp, ss;
} bios_reg_t;

bios_reg_t MYREGS = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

//#endif

/****************************************************************************/
/*                                                                          */
/* Local Prototypes                                                         */
/*                                                                          */
/****************************************************************************/
/****************************************************************************/
/*                                                                          */
/* Define macros to obtain individual bytes from a word register            */
/*                                                                          */
/****************************************************************************/

#define HIGH_BYTE(ax) (ax >> 8)
#define LOW_BYTE(ax) (ax & 0xff)

/****************************************************************************/
/*                                                                          */
/* PROT_GENINTERRUPT                                                        */
/*                                                                          */
/****************************************************************************/
extern void
pci_geninterrupt (void)
{
#ifndef EFI
	union REGS regs;
	struct SREGS sregs;

	memset(&sregs, 0, sizeof(sregs));
	regs.w.ax = 0x300;
	regs.h.bl = 0x1a;
	regs.h.bh = 0;
	regs.w.cx = 0;
	sregs.es = FP_SEG(&MYREGS);
	regs.x.edi = FP_OFF(&MYREGS);
	int386x (0x31, &regs, &regs, &sregs);
#endif
}

/****************************************************************************/
/*                                                                          */
/* PCI_BIOS_PRESENT                                                         */
/*                                                                          */
/* Purpose: Determine the presence of the PCI BIOS                          */
/*                                                                          */
/* Inputs: None                                                             */
/*                                                                          */
/* Outputs:                                                                 */
/*                                                                          */
/*   u08 *hardware_mechanism                                               */
/*      Identifies the hardware characteristics used by the platform.       */
/*      Value not assigned if NULL pointer.                                 */
/*         Bit 0 - Mechanism #l supported                                   */
/*         Bit 1 - Mechanism #2 supported                                   */
/*                                                                          */
/*   u16 *interface_level_version                                          */
/*     PCI BIOS Version - Value not assigned if NULL pointer.               */
/*             High Byte - Major version number                             */
/*             Low Byte - Minor version number                              */
/*                                                                          */
/*   u08 *last_pci_bus_number                                            */
/*   Number of last PCI bus in the system. Value not assigned if NULL       */
/*      pointer                                                             */
/*                                                                          */
/*   Return Value - Indicates presence of PCI BIOS                          */
/*      SUCCESSFUL - PCI BIOS Present                                       */
/*      NOT_SUCCESSFUL - PCI BIOS Not Present                               */
/*                                                                          */
/****************************************************************************/

int pci_bios_present(u08 *hardware_mechanism,
		     u16 *interface_level_version,
		     u08 *last_pci_bus_number)
{

   int ret_status;          /* Function Return Status.                      */
   u08 bios_present_status;/* Indicates if PCI bios present                */
   u32 pci_signature;     /* PCI Signature ('P', 'C', 'I', ' ')           */
   u16 ax, bx, cx, flags;  /* Temporary variables to hold register values  */

   /* Call PCI BIOS Int 1Ah interface */

   MYREGS.eax = (PCI_FUNCTION_ID << 8) | PCI_BIOS_PRESENT;
   /*MYREGS.flags = _FLAGS;*/
   pci_geninterrupt ();

   /* Save registers before overwrltten by compiler usage of registers */
   ax = MYREGS.eax & 0xffff;
   bx = MYREGS.ebx & 0xffff;
   cx = MYREGS.ecx & 0xffff;
   pci_signature = MYREGS.edx;
   flags = MYREGS.flags;
   bios_present_status = HIGH_BYTE(ax);

   /* First check if CARRY FLAG Set, if so, BIOS not present */
   if ((flags & CARRY_FLAG) == 0) {

      /* Next, must check that AH (BIOS Present Status) == 0 */
      if (bios_present_status == 0) {

	 /* Check bytes in pci_signature for PCI Signature */
	 if ((pci_signature & 0xff)     == 'P' &&
	 ((pci_signature >> 8) & 0xff)  == 'C' &&
	 ((pci_signature >> 16) & 0xff) == 'I' &&
	 ((pci_signature >> 24) & 0xff) == ' ') {

	    /* Indicate to caller that PCI bios present */
	    ret_status = SUCCESSFUL;

	    /* Extract calling parameters from saved registers */
	    if (hardware_mechanism != NULL) {
		*hardware_mechanism = LOW_BYTE(ax);
	    }
	    if (hardware_mechanism != NULL) {
		*interface_level_version = bx;
	    }
	    if (hardware_mechanism != NULL) {
		*last_pci_bus_number = LOW_BYTE(cx);
	    }
	 }
      }
      else {
      ret_status = NOT_SUCCESSFUL;
      }
   }
   else {
      ret_status = NOT_SUCCESSFUL;
   }

   return (ret_status);

} 


/****************************************************************************/
/*                                                                          */
/*  FIND_PCI_DEVICE                                                         */
/*                                                                          */
/* Purpose: Determine the location of PCI devices given a specific Vendor   */
/*          Device ID and Index number.  To find the first device, specify  */
/*          0 for index, 1 in index finds the second device, etc.           */
/*                                                                          */
/* Inputs:                                                                  */
/*                                                                          */
/*    u16 device_id                                                         */
/*       Device ID of PCI device desired                                    */
/*                                                                          */
/*    u16 vendor_id                                                         */
/*       Vendor ID of PCI device desired                                    */
/*                                                                          */
/*    u16 index                                                             */
/*       Device number to find (0 - (N-1))                                  */
/*                                                                          */
/* Outputs:                                                                 */
/*                                                                          */
/*    u08 *bus_number                                                       */
/*       PCI Bus in which this device is found                              */
/*                                                                          */
/*    u08 *device_and_function                                              */
/*       Device Number in upper 5 bits, Function Number in lower 3 bits     */
/*                                                                          */
/*    Return Value - Indicates presence of device                           */
/*       SUCCESSFUL - Device found                                          */
/*       NOT_SUCCESSFUL - BIOS error                                        */
/*       DEVICE_NOT_FOUND - Device not found                                */
/*       BAD_VENDOR_ID - Illegal Vendor ID (0xffff)                         */
/*                                                                          */
/****************************************************************************/

int find_pci_device(u16 device_id,
		    u16 vendor_id,
		    u16 index,
		    u08 *bus_number,
		    u08 *device_and_function)
{

  int ret_status;     /* Function Return Status                      */
  u16 ax, bx, flags; /* Temporary variables to hold register values */

  /* Load entry registers for PCI BIOS */
  MYREGS.ecx = device_id;
  MYREGS.edx = vendor_id;
  MYREGS.esi = index;
  MYREGS.eax = (PCI_FUNCTION_ID << 8) | FIND_PCI_DEVICE;
  pci_geninterrupt ();

  /* Save registers before overwritten by compiler usage of registers */
  ax = MYREGS.eax & 0xffff;
  bx = MYREGS.ebx & 0xffff;
  flags = MYREGS.flags;

  /* First check if CARRY FLAG Set, if so, error has occurred */
  if ((flags & CARRY_FLAG) == 0){
     /* Get Return code from BIOS */
     ret_status = HIGH_BYTE(ax);
     if (ret_status == SUCCESSFUL) 
	 {
		/* Assign Bus Number, Device & Function if successful */

		if (bus_number != NULL) 
		   *bus_number = HIGH_BYTE(bx);
	
		if (device_and_function != NULL) 
		   *device_and_function = LOW_BYTE(bx);
     }
  }
  else {
  ret_status = NOT_SUCCESSFUL;
  }
  return (ret_status);

}




/****************************************************************************/
/*                                                                          */
/*  FIND_PCI_CLASS_CODE                                                     */
/*                                                                          */
/* Purpose: Returns the location of PCI devices that have a specific Class  */
/*           Code.                                                          */
/*                                                                          */
/* Inputs:                                                                  */
/*                                                                          */
/*   u16 class_code                                                        */
/*      Class Code of PCI device desired                                    */
/*                                                                          */
/*   u16 index                                                             */
/*      Device number to find (0 - (N-l))                                   */
/*                                                                          */
/* Outputs:                                                                 */
/*                                                                          */
/*   u08 *bus_number                                                       */
/*      PCI Bus in which this device is found                               */
/*                                                                          */
/*   u08 *device_and_function                                              */
/*      Device Number in upper 5 bits, Function Number in lower 3 bits      */
/*                                                                          */
/*   Return Value - Indicates presence of device                            */
/*      SUCCESSFUL - Device found                                           */
/*      NOT_SUCCESSFUL - BIOS error                                         */
/*      DEVICE_NOT_FOUND - Device not found                                 */
/*                                                                          */
/****************************************************************************/

int find_pci_class_code(u32 class_code,
			u16 index,
			u08 *bus_number,
			u08 *device_and_function)

{
   int ret_status;     /* Function Return Status                      */
   u16 ax, bx, flags; /* Temporary variables to hold register values */

   /* Load entry registers for PCI BIOS */
   MYREGS.ecx = class_code;
   MYREGS.esi = index;
   MYREGS.eax = (PCI_FUNCTION_ID << 8) | FIND_PCI_CLASS_CODE;
   pci_geninterrupt ();

   /* Save registers before overwritten by compiler usage of registers */
   ax = MYREGS.eax & 0xffff;
   bx = MYREGS.ebx & 0xffff;
   flags = MYREGS.flags;

   /* First check if CARRY FLAG Set, if so, error has occurred */
   if ((flags & CARRY_FLAG) == 0) {

      /* Get Return code from BIOS */
      ret_status = HIGH_BYTE(ax);
      if (ret_status == SUCCESSFUL) {

	 /* Assign Bus Number, Device & Function if successful */
	 if (bus_number != NULL) {
	    *bus_number = HIGH_BYTE(bx);
	 }

	 if (device_and_function != NULL) {
	    *device_and_function = LOW_BYTE(bx);
	 }

      }

   }
   else {
      ret_status = NOT_SUCCESSFUL;
   }

   return (ret_status);

   /* EFI */
} 


/****************************************************************************/
/*                                                                          */
/*  READ_CONFIGURATION_BYTE                                                 */
/*                                                                          */
/* Purpose: Reads a byte from the configuration space of a specific device  */
/*                                                                          */
/* Inputs:                                                                  */
/*                                                                          */
/*   u08 bus_number                                                        */
/*      PCI bus to read configuration data from                             */
/*                                                                          */
/*   u08 device_and_function                                               */
/*      Device Number in upper 5 bits, Function Number in lower 3 bits      */
/*                                                                          */
/*   u08 register_number                                                   */
/*      Register Number of configuration space to read                      */
/*                                                                          */
/* Outputs:                                                                 */
/*                                                                          */
/*   u08 *byte_read                                                         */
/*      Byte read from Configuration Space                                  */
/*                                                                          */
/*   Return Value - Indicates presence of device                            */
/*      SUCCESSFUL - Device found                                           */
/*      NOT_SUCCESSFUL - BIOS error                                         */
/*      BAD_REGISTER_NUMBER - Invalid Register Number                       */
/*                                                                          */
/****************************************************************************/

int read_configuration_byte(u08 bus_number,
			    u08 device_and_function,
			    u08 register_number,
			    u08 *byte_read)
{
    if (NOCFG() && (register_number > 0x30))
	{
        *byte_read = t_readmem08(register_number);
        *byte_read = *(u08*) (CURRCARD.base_address + register_number);
	}
    else


    {
        u08 value = 0;
        u16 data_addr, cmd_addr;
        u32 cmd;

        cmd_addr = 0xcf8;
        cmd = CONFIG_CMD(bus_number,device_and_function,register_number);
        data_addr = 0xcfc+(register_number & 3);

        __asm {
			pushf
			cli
			mov dx, cmd_addr
			mov eax,cmd
			out	dx, eax
			mov dx, data_addr
			in  al, dx
			mov value, al
			popf
		} /* __asm */

        *byte_read = value;
        iprintf("PCI cfg rd: %02X --> %02X\n",register_number, *byte_read);
    }    

    return SUCCESSFUL;
} 


/****************************************************************************/
/*                                                                          */
/*  READ_CONFIGURATION_WORD                                                 */
/*                                                                          */
/* Purpose: Reads a u16 from the configuration space of a specific device  */
/*                                                                          */
/* Inputs:                                                                  */
/*                                                                          */
/*   u08 bus_number                                                        */
/*      PCI bus to read configuration data from                             */
/*                                                                          */
/*   u08 device_and_function                                               */
/*      Device Number in upper 5 bits, Function Number in lower 3 bits      */
/*                                                                          */
/*   u08 register_number                                                   */
/*      Register Number of configuration space to read                      */
/*                                                                          */
/* Outputs:                                                                 */
/*                                                                          */
/*   u16 *word_read                                                        */
/*      Word read from Configuration Space                                  */
/*                                                                          */
/*   Return Value - Indicates presence of device                            */
/*      SUCCESSFUL - Device found                                           */
/*      NOT_SUCCESSFUL - BIOS error                                         */
/*      BAD_REGISTER_NUMBER - Invalid Register Number                       */
/*                                                                          */
/****************************************************************************/ 

int read_configuration_word(u08 bus_number,
			    u08 device_and_function,
			    u08 register_number,
			    u16 *word_read)
{
    
    {
        u16 data_addr, cmd_addr, value = 0;
        u32 cmd;

        cmd_addr = 0xcf8;
        cmd = CONFIG_CMD(bus_number,device_and_function,register_number);
        data_addr = 0xcfc+(register_number & 2);


        __asm {
			pushf
			cli
			mov dx, cmd_addr
			mov eax,cmd
			out	dx, eax
			mov dx, data_addr
			in  ax, dx
			mov value, ax
			popf
		} /* __asm */

        *word_read = value;
        iprintf("PCI cfg rd: %02X --> %04X\n",register_number, *word_read);
    }    
//#endif
    return SUCCESSFUL;
}


/****************************************************************************/ 
/*                                                                          */
/* READ_CONFIGURATION_DWORD                                                 */
/*                                                                          */
/* Purpose: Reads a u32 from the configuration space of a specific device */
/*                                                                          */
/* Inputs:                                                                  */
/*                                                                          */
/*   u08 bus_number                                                        */
/*      PCI bus to read configuration data from                             */
/*                                                                          */
/*   u08 device_and_function                                               */
/*      Device Number in upper 5 bits, Function Number in lower 3 bits      */
/*                                                                          */
/*   u08 register_number                                                   */
/*      Register Number of configuration space to read                      */
/*                                                                          */
/* Outputs:                                                                 */
/*                                                                          */
/*   u32 *dword_read                                                      */
/*      Dword read from Configuration Space                                 */
/*                                                                          */
/*   Return Value - Indicates presence of device                            */
/*      SUCCESSFUL - Device found                                           */
/*      NOT_SUCCESSFUL - BIOS error                                         */
/*      BAD_REGISTER_NUMBER - Invalid Register Number                       */
/*                                                                          */
/****************************************************************************/ 

int read_configuration_dword(u08 bus_number,
			     u08 device_and_function,
			     u08 register_number,
			     u32 *dword_read)
{
    {
        u16 data_addr, cmd_addr;
        u32 value = 0, cmd;

        cmd_addr = 0xcf8;
        cmd = CONFIG_CMD(bus_number,device_and_function,register_number);
        data_addr = 0xcfc;

        //outpd(0xcf8,CONFIG_CMD(bus_number,device_and_function,register_number));
        //*dword_read = inpd(0xcfc);
  
        __asm {
			pushf
			cli
			mov dx, cmd_addr
			mov eax,cmd
			out	dx, eax
			mov dx, data_addr
			in  eax, dx
			mov value, eax
			popf
		} /* __asm */

        *dword_read = value;
        iprintf("PCI cfg rd: %02X --> %08X\n",register_number, *dword_read);
    }    

    return SUCCESSFUL;
}

/****************************************************************************/
/*                                                                          */
/*  WRITE_CONFIGURATION_BYTE                                                */
/*                                                                          */
/* Purpose: Writes a u08 to the configuration space of a specific device   */
/*                                                                          */
/* Inputs:                                                                  */
/*                                                                          */
/*   u08 bus_number                                                        */
/*      PCI bus to write conflguration data to                              */
/*                                                                          */
/*   u08 device_and_function                                               */
/*      Device Number in upper 5 bits, Function Number in lower 3 bits      */
/*                                                                          */
/*   u08 register_number                                                   */
/*      Register Number of configuration space to write                     */
/*                                                                          */
/*   u08 byte_to_write                                                     */
/*      Byte to write to Configuration Space                                */
/*                                                                          */
/* Outputs:                                                                 */
/*                                                                          */
/*   Return Value - Indicates presence of device                            */
/*      SUCCESSFUL - Device found                                           */
/*      NOT_SUCCESSFUL - BIOS error                                         */
/*      BAD_REGISTER_NUMBER - Invalid Register Number                       */
/*                                                                          */
/****************************************************************************/

int write_configuration_byte(u08 bus_number,
			     u08 device_and_function,
			     u08 register_number,
			     u08 byte_to_write)
{
   {
        u08 value = byte_to_write;
        u16 data_addr, cmd_addr;
        u32 cmd;

        iprintf("PCI cfg wr: %02X <-- %02X\n",register_number, byte_to_write);
        cmd_addr = 0xcf8;
        cmd = CONFIG_CMD(bus_number,device_and_function,register_number);
        data_addr = 0xcfc+(register_number & 3);

        //outpd(0xcf8,CONFIG_CMD(bus_number,device_and_function,register_number));
        //outp(0xcfc+(register_number&3),byte_to_write);

        __asm {
			pushf
			cli
			mov dx, cmd_addr
			mov eax,cmd
			out	dx, eax
			mov dx, data_addr
			mov al, value
			out dx, al
			popf
		} /* __asm */
    }    

    return SUCCESSFUL;
}


/****************************************************************************/
/*                                                                          */
/* WRITE_CONFIGURATION_WORD                                                 */
/*                                                                          */
/* Purpose: Writes a u16 to the configuration space of a specific device    */
/*                                                                          */
/* Inputs:                                                                  */
/*                                                                          */
/*   u08 bus_number                                                         */
/*      PCI bus to read configuration data from                             */
/*                                                                          */
/*   u08 device_and_function                                                */
/*      Device Number in upper 5 bits, Function Number in lower 3 bits      */
/*                                                                          */
/*   u08 register_number                                                    */
/*      Register Number of configuration space to read                      */
/*                                                                          */
/*   u16 word_to_write                                                      */
/*      Word to write to Configuration Space                                */
/*                                                                          */
/* Outputs:                                                                 */
/*                                                                          */
/*   Return Value - Indicates presence of device                            */
/*      SUCCESSFUL - Device found                                           */
/*      NOT_SUCCESSFUL - BIOS error                                         */
/*      BAD_REGISTER_NUMBER - Invalid Register Number                       */
/*                                                                          */
/****************************************************************************/

int write_configuration_word(u08 bus_number,
			     u08 device_and_function,
			     u08 register_number,
			     u16 word_to_write)

{
	
    {
        u32 cmd;
        u16 value = word_to_write;
        u16 data_addr, cmd_addr;

        iprintf("PCI cfg wr: %02X <-- %04X\n",register_number, word_to_write);
        cmd_addr = 0xcf8;
        cmd = CONFIG_CMD(bus_number,device_and_function,register_number);
        data_addr = 0xcfc+(register_number & 2);

        //outpd(0xcf8,CONFIG_CMD(bus_number,device_and_function,register_number));
        //outpw(0xcfc+(register_number&2),word_to_write);
  
        __asm {
			pushf
			cli
			mov dx, cmd_addr
			mov eax,cmd
			out	dx, eax
			mov dx, data_addr
			mov ax, value
			out dx, ax
			popf
		} /* __asm */
    }    

    return (SUCCESSFUL);
} 


/****************************************************************************/
/*                                                                          */
/*  WRITE_CONFIGURATION_DWORD                                               */
/*                                                                          */
/* Purpose: Reads a u32 from the configuration space of a specific device */
/*                                                                          */
/* Inputs:                                                                  */
/*                                                                          */
/*   u08 bus_number                                                        */
/*      PCI bus to read configuration data from                             */
/*                                                                          */
/*   u08 device_and_function                                               */
/*      Device Number in upper 5 bits, Function Number in lower 3 bits      */
/*                                                                          */
/*   u08 register_number                                                   */
/*      Register Number of configuration space to read                      */
/*                                                                          */
/*   u32 dword_to_write                                                   */
/*      Dword to write to Configuration Space                               */
/*                                                                          */
/* Outputs:                                                                 */
/*                                                                          */
/*   Return Value - Indicates presence of device                            */
/*      SUCCESSFUL - Device found                                           */
/*      NOT_SUCCESSFUL - BIOS error                                         */
/*      BAD_REGISTER_NUMBER - Invalid Register Number                       */
/*                                                                          */
/****************************************************************************/ 
int write_configuration_dword(u08 bus_number,
			      u08 device_and_function,
			      u08 register_number,
			      u32 dword_to_write)

{
    {
        u32 cmd, value = dword_to_write;
        u16 data_addr, cmd_addr;

        iprintf("PCI cfg wr: %02X <-- %08X\n",register_number, dword_to_write);
        cmd_addr = 0xcf8;
        cmd = CONFIG_CMD(bus_number,device_and_function,register_number);
        data_addr = 0xcfc;
  
        __asm {
			pushf
			cli
			mov dx, cmd_addr
			mov eax,cmd
			out	dx, eax
			mov dx, data_addr
			mov eax, value
			out dx, eax
			popf
		} /* __asm */
    }    

    return (SUCCESSFUL);
}

