/********************************************************************
 * Copyright(c) 2000 Broadcom Corporation, all rights reserved.
 * Contains proprietary and confidential information.
 *
 * This source file is the property of Broadcom Corporation, and
 * may not be copied or distributed in any isomorphic form without 
 * the prior written consent of Broadcom Corporation. 
 *
 *  FileName: $Header: /Source/Bcm44xx/b44diag/t3_cli.c 9     7/19/02 10:07a Ahui $
 *
 *  Description: This file contains command line interface 
 *
 *  $Log: /Source/Bcm44xx/b44diag/t3_cli.c $ 
 * 
 * 9     7/19/02 10:07a Ahui
 * Removed unused code
 * 
 * 8     6/11/02 3:55p Ahui
 * Modified "?" message display
 * 
 * 7     4/25/02 11:08a Ahui
 * 
 * 3     3/07/02 5:05p Ahui
 * Added sromutil support
 ********************************************************************/ 

#include <ctype.h>

#include "general.h"
#include "t3_cli.h"
#include "t3_hist.h"
#include "t3_gdata.h"
#include "t3_tests.h"
#include "t3_scrn.h"
#include "t3_win.h"
#include "t3_comm.h"

#define COMMAND_LENGTH MAX_LINE_LENGTH

/*********** static variables ************/

static s08 prompt[MAX_PROMPT_LENGTH+1]=">";
                    
/****************************************************************************
 ****************************************************************************
                        private routines 
 ****************************************************************************
 ****************************************************************************/
 
/*
 * CMD_optionBad
 *      validate the option. If valid, update cmdPtr option values
 *
 * input cmdPtr - command structure pointer
 *       str - parameter input
 *
 * output - modifies cmdPtr option values
 *
 * return 1 - bad option 
 *        0 - good option
 */
static int CMD_optionBad(command_ptr cmdPtr, s08 * str)
{
    option_ptr optPtr;
    
    if (cmdPtr->opt == NULL)
        return 1; /* no option available */
        
    while (*str)
    {
        optPtr = cmdPtr->opt;
        while (optPtr->key)
        {
            if (optPtr->key == *str) /* found value option */
            {
                optPtr->entered++;
                if (*(++str)=='=') str++;
                switch (optPtr->type) 
                {
                    case OPT_U32:
                        optPtr->val.value32 = strtoul(str, &str, 16);
                        break;
                    case OPT_STR:
                        optPtr->val.valueStr = str;
                        return 0;
                    case OPT_INT:
                        optPtr->val.valueInt = strtol(str, &str, 16);
                        break;
                    case OPT_BOOL:
                        optPtr->val.value32 = !optPtr->val.value32;
                        break;
                    case OPT_DEC:
                    case OPT_LOOP:
                        optPtr->val.valueInt = strtol(str, &str, 10);
                        break;
                    default:
                        return 1; /* type not supported */
                }
                break;
            }
            optPtr++;
        }
        if (!optPtr->key) /* option not supported */
            return 1;
    }
    return 0; /* good option */
}

/*
 * CMD_checkOption
 *      check for command line option and remove the options from argc,argu
 *
 * input cmd - command structure pointer
 *       argc - argument count
 *       argu - argument list
 * 
 * output argc - modified new argument count
 *        argu - modified new argument list
 *
 * return error code
 */
static int CMD_checkOption(command_ptr cmd, int * argc,s08 **argu)
{
    int para, index = 0, cnt = *argc;
    s08 * token;
    
    cmd_clearOption(cmd);
    while (index < cnt)
    {   
        token = argu[index++];
        if (*token=='?')
        {
            cmd_help(cmd);
            return ERR_INVALID;
        }
        if (*token++ == '-') /* it's option */
        {
            if (*token == '?') /* help */
            {
                cmd_help(cmd);
                return ERR_INVALID;
            }
            
            if (CMD_optionBad(cmd, token))
            {
                return test_failed0(ERR_OPTION);
            }
            
            /* move up parameter and overwrite option */
            index--;
            cnt--;
            (*argc)--;
            for (para = index; para < cnt; para++)
                argu[para] = argu[para+1];
        }
    }
    return NO_ERROR;
}

/* 
 *
 */
static int find_exact_cmd(s08 * cmdStr, command_ptr * cmdPtr, command_ptr * cmdSet)
{
    command_ptr * findCmds;

    if (cmdSet != NULL)
        for (findCmds = cmdSet; *findCmds; findCmds++)
            if (!strcmp(cmdStr, (*findCmds)->cmd))
            {
                *cmdPtr = *findCmds;
                return NO_ERROR;
            }
    return HAS_ERROR;
}
 
/* 
 *
 */
static int find_partial_cmd(s08 * cmdStr, command_ptr * cmdPtr, command_ptr * cmdSet)
{
    int len;
    command_ptr * findCmds;
 
    if (cmdSet == NULL)
        return HAS_ERROR;
        
    for (findCmds=cmdSet, len=strlen(cmdStr); *findCmds; findCmds++)
    {
        if (!strncmp(cmdStr, (*findCmds)->cmd, len)) /* found one */
        {
            for (*cmdPtr = *findCmds++; *findCmds; findCmds++)
            {
                if (!strncmp(cmdStr, (*findCmds)->cmd, len)) /* found another */        
                {
                    /* print out that we found so far */
                    printf("%s, %s",(*cmdPtr)->cmd, (*findCmds++)->cmd);
                    
                    /* print out any others if there is more */
                    for (;*findCmds;findCmds++) 
                        if (!strncmp(cmdStr, (*findCmds)->cmd, len))
                            printf(", %s",(*findCmds)->cmd);
                            
                    printf(" ???\nAmbiguous command\n");
                    return HAS_ERROR;
                }
            }
            return NO_ERROR;
        }
    }
    return HAS_ERROR;
}
 
/* 
 * CMD_getCommand(s08 * cmd, command_ptr * cmds)
 *      find the supported command. return NULL if the command is not valid.
 *      the command can be entered only first few characters if there is no
 *      ambigous command shares the same string
 *
 *      in case of ambiguous or invalid command, the message is printed
 *
 * input: cmdStr - input cmd
 *
 * output: cmdPtr - upon successful return, it points to command structure
 *
 * return: error code
 */
static int CMD_getCommand(s08 * cmdStr, command_ptr * cmdPtr, command_ptr * extCmdSet)
{
    int status;
    
    /* find exact match first from external command set */
    if (find_exact_cmd(cmdStr, cmdPtr, extCmdSet)==NO_ERROR)
        return NO_ERROR; 
    /* find exact match first from default command set */
    if (find_exact_cmd(cmdStr, cmdPtr, cmds)==NO_ERROR)
        return NO_ERROR; 

    /* find partial match from external command set */
    if ((status=find_partial_cmd(cmdStr, cmdPtr, extCmdSet))!=HAS_ERROR)
        return status;
    
    /* find partial match from default command set */
    if ((status=find_partial_cmd(cmdStr, cmdPtr, cmds))!=HAS_ERROR)
        return status;
    
    eprintf("Unknown command '%s'\n",cmdStr);
    eprintf("For help, type 'help%s'\n\n",extCmdSet==NULL?" <command group>":"");
    cmd_help(&cmd_help1);
    return ERR_COMMAND;
}
 
/* 
 * CMD_loop 
 *
 * input - argc, argu
 *
 * return - exit Code
 */
int CMD_loop(s08 * buff, command_ptr * cmdSet)
{
    s08 loop_buff[COMMAND_LENGTH+1];
    u32 loop = 0;
    u32 loopCnt = 0;
    int retc = NO_ERROR;
    
    while (*buff==' ') buff++;
    
    if (*buff=='?' || *buff=='-')
        return cmd_help(&cmd_loop);

    if (util_isDigit(*buff))
        loop = strtoul(buff,&buff,10);

    do 
    {
        win_setTextColor(BROWN);
        printf("\nLoop %d",++loopCnt);
        if (loop)
            printf(" of %d",loop);
        printf("\n");
        win_setTextColor(GRAY);
        strcpy(loop_buff,buff);
    } while (((retc=cmd_executeString(loop_buff, cmds))==NO_ERROR) && 
             !util_user_abort() &&
             (loopCnt<loop || !loop));
    return retc;
}


#ifndef WIN_CLI
/*
 * CMD_getprompt
 *    get a cmmand input line 
 * 
 * input -- buff, buffer for input line
 *
 */ 
static void CMD_getCommandPrompt(s08 * buff)
{
    if (global.G_ifile)  
    {
        /* read command from input file */ 
        if (global.verbose)
            printf("%s",prompt);
        if (fgets (buff, COMMAND_LENGTH, global.G_ifile))
        {
            if (global.verbose)
                printf("%s\n",buff);
            return;
        }
        fclose (global.G_ifile);
        global.G_ifile = NULL;
    }
    /* get command from keyboard */
    printf("%s",prompt);
    strcpy(buff, get_line(util_poll_cpu_messages));
}

/*
 * checkForAtOption
 *    if @ is entered in command line, run all command in the specified device;
 * 
 * return HAS_ERROR if invliad device is specified
 *        NO_ERROR if no error detected
 *        global.devSelect bits will indicate devices
 *
 */ 
static int checkForAtOption(s08 * buff,s08 ** nextStr)
{
    u32 data;
    
    global.devSelect = 0;
    *nextStr = NULL;
    
    while (*buff && *buff!='@')
        buff++;
        
    *nextStr = strchr(buff,';');
    if (*nextStr!=NULL)
        *(*nextStr)++ = '\0';
    
    if (*buff)
    {
        *buff++ = '\0';
        while (*buff==' ')
            buff++;
        if (!(*buff && stricmp(buff,"all")))
            global.devSelect = ALL_DEV_BIT;
        else
        {
            while (*buff)
            {
                while (*buff && !isdigit(*buff))
                    buff++;
                if (*buff)
                {
                    data = strtoul(buff,&buff,10);
                    if (data >= (u32) global.device_detected)
                    {
                        eprintf("Invalid device %d\n",data);
                        return HAS_ERROR;
                    }
                    global.devSelect |= (1 << data);
                }
            }
        }        
    }
    return NO_ERROR;
}

/*****************************************************************************/
int executeSelect(int * (func)(void))
{
    u32 bits = global.blastCntl;
    u32 dev;
    u8  oldDev = global.current_card;

    for (dev=0;bits;dev++,bits>>=1)
    {
        if (bits & 1)
        {
            global.current_card = dev;
            if (func()!=NO_ERROR)
            {
                global.current_card = oldDev;
                return HAS_ERROR;
            }
        }
    }
    global.current_card = oldDev;
    return NO_ERROR;            
}

/****************************************************************************/
static int CMD_Iteration(command_ptr pCmd)
{
    option_ptr optr = pCmd->opt;
    
    if (optr!=NULL)
    {
        while (optr->key)
        {
            if (optr->type==OPT_LOOP)
                return optr->entered ? optr->val.valueInt : 1;
            optr++;
        }
    }
    return 1; 
}

static int runMultiDevCommand(command_ptr cmd, int argc,s08 *argu[], s08 * nextOpt)
{
    u08 dev;
    u08 orgDev = global.current_card;
    u32 select;
    u32 blastSel = 0;
    int retc = NO_ERROR;
    u08 * buff;

    com_disable();
    while (retc==NO_ERROR)
    {
        select =  global.devSelect ? global.devSelect : 1 << global.current_card;
        blastSel |= select;
        for (dev=0; select && (dev<global.device_detected); dev++)
        {
            if (select & 1)
            {
                if (!util_switch_card(dev))
                {   
                    retc = HAS_ERROR;
                    break;
                }    
                if ((retc = (!strcmp(cmd->cmd,"blast") ? blastInitCfg(cmd) : txInitCfg(cmd)))!=NO_ERROR)
                    break;
            }
            select >>= 1;
        }
        if (retc==NO_ERROR)
        {
            util_switch_card(orgDev);
            if (nextOpt==NULL)
                break;
            buff = nextOpt;
            if ((retc = checkForAtOption(buff,&nextOpt)) != NO_ERROR)
                break;
            cmd_str2Parameter(buff, &argc, argu);
            if ((retc = CMD_checkOption(cmd, &argc, argu))==ERR_OPTION)
                break;
        }    
    }
    if (retc==NO_ERROR)
    {
        global.devSelect = blastSel;
        retc = cmd->func(argc, argu);
    }    
    com_enable();
    return retc;
}

static int runSingleDevCommand(command_ptr cmd, int argc,s08 *argu[])
{
    u08 dev;
    u08 orgDev = global.current_card;
    u32 select;
    u32 dispDev = global.devSelect;
    int retc = NO_ERROR;
    int iteration,iter;
    
    iteration = CMD_Iteration(cmd);
    iter=0;

    do 
    {
        showIteration(++iter,iteration);
        if (iteration !=1)
            printf("\n");
        if (!global.devSelect)
            retc = cmd->func(argc, argu);
        else
        {
            select =  global.devSelect;
            for (dev=0;select && (dev<global.device_detected) && (retc==NO_ERROR); dev++)
            {
                if (select & 1)
                {
                    if (!util_switch_card(dev))
                        break;
                    if (dispDev)
                    {
                        win_setTextColor(GREEN);
                        printf("Device %d\n",dev);
                        win_setTextColor(GRAY);
                    }    
                    retc = cmd->func(argc, argu);
                    if (util_user_abort())
                        retc = ERR_ABORT;
                }    
                select >>= 1;
            }
        }
        if (util_user_abort())
            break;
    } while ((!iteration || (iter < iteration)) && (retc==NO_ERROR));
    if (strcmp(cmd->cmd,"device"))
        util_switch_card(orgDev);
        
    return retc;
}

/*
 * runCommand
 *    if @ is entered in command line, run all command in the specified device;
 * 
 * input -- buff, buffer for input line
 *
 */ 
static int runCommand(command_ptr cmd,int argc,s08 *argu[], s08 * nextOpt)
{
    if (!strcmp(cmd->cmd,"blast") || !strcmp(cmd->cmd,"txpkt"))
        return runMultiDevCommand(cmd, argc, argu, nextOpt);
    else
        return runSingleDevCommand(cmd, argc, argu);
}

/****************************************************************************
 ****************************************************************************
                        public routines 
 ****************************************************************************
 ****************************************************************************/
/*****************************************************************************/

/* 
 * cmd_executeString
 * 
 * input - command string
 *         cmd set
 *
 * return - exit Code
 */
int cmd_executeString(s08 * buff, command_ptr * cmds)
{
    int argc;
    s08 * argu[MAX_PARAMETER];
    command_ptr cmd;
    int retc = NO_ERROR;
    s08 * nextOpt;

    /* if log file is opened, we want to flush out all written data to disk first just in case
       if the following command hang the system and lose all the log content */
	if (global.logfile!=NULL)
    {
        /* fflush(global.logfile);  -- flush didn't work 
          So, we close the file and reopen again with append */
        fclose(global.logfile);                               
        global.logfile = fopen(global.logFileName, "a+");                                       
    }    
    while (*buff==' ') buff++; 
#ifndef EFI
    if (!strnicmp(buff,"dos ",4) || (!strnicmp(buff,"dos",3) && (strlen(buff)==3)))
    {
        buff += 3;
        while (*buff==' ') buff++; 
		if (strlen(buff))
        {
            if (buff[0]=='?' || buff[1]=='?')
                cmd_help(&cmd_dos);
            else
            {
                system (buff);
                win_gotoxy( __scrn.rsize-1,0);
                printf("\n");
            }    
        }
		else
		{
            system ("command.com");
            util_print_banner();
		}
    }
    else
#endif
    if (!strnicmp(buff,"loop ",5))
    {
        retc = CMD_loop(buff+5,cmds);
    }
    else if (!strnicmp(buff,"echo ",5))
    {
        printf("%s\n",buff+5);
    }
    else
    {
        if ((retc = checkForAtOption(buff,&nextOpt)) != HAS_ERROR)
        {
            cmd_str2Parameter(buff, &argc, argu);
            if (argc)
            {
                if (*buff == '?' || (*buff=='-' && *(buff+1)=='?'))
					return runCommand(&cmd_help1, argc, argu, nextOpt);
				else
				{
					if ((retc = CMD_getCommand(argu[0],&cmd, cmds)) == NO_ERROR)
						retc = CMD_checkOption(cmd, &argc, argu) ? ERR_OPTION :
							  (cmd->func == NULL ? CMD_QUIT : runCommand(cmd, argc, argu, nextOpt));
				}
            }
        }   
    }
    return retc;
}  

/*
 * cmd_str2Parameter 
 *      convert input string to C parameter style argc and argu
 *
 * input: buff - input buffer string
 *
 * output: argc - argument acount
 *         argu - argument pointer
 */
void cmd_str2Parameter(s08 * buff, int * argc, s08 **argu)
{
    *argc = 0;
    while (*buff && (*argc < MAX_PARAMETER))
    {
        while (*buff==' ') buff++; /* skip spaces */
        if (*buff)
        {
            *argu++ = buff;
            (*argc)++;
            while (*buff && *buff!=' ') buff++; /* skip to the end of token */
            if (*buff) *buff++ = '\0';  /* terminate string */
        }
    }
}

/****************************************************************************/
/****************************************************************************/
void cmd_setPrompt(s08 * newPrompt)
{
    if (strlen(newPrompt) <= MAX_PROMPT_LENGTH)
        strcpy(prompt, newPrompt);
    else
        eprintf("Requested new prompt is too long, prompt not changed\n");
}

/****************************************************************************/
int cmd_getPrompt(s08 * buffer, u32 len)
{
    if (len <= strlen(prompt))
    {
        eprintf("No enough space to get prompt string\n");
        return 0;
    }
    strcpy(buffer, prompt);
    return 1;
}

#endif
 
/****************************************************************************
 * cmd_line_interface()
 *    It prompt user the enter the command, and execute as nessary. The 
 *    concept is similar to a DOS prompt
 * 
 * input - command string. If the command string is NULL, it will 
 *         stay in the loop to get the prompt and excute. Otherwise, it use 
 *         the input string as command and only excute that parameter 
 * 
 * output - none
 * 
 * return - none
 **************************************************************************/
int cmd_line_interface(s08 * para)
{        
#ifndef WIN_CLI
    s08 buff[COMMAND_LENGTH+1];
    if (para != NULL)
#endif    
        return cmd_executeString(para, NULL);
#ifndef WIN_CLI
//    global.CmdMode = 1;
    do CMD_getCommandPrompt(buff);
    while (cmd_executeString(buff, NULL) != CMD_QUIT);
//    global.CmdMode = 0;
    return NO_ERROR;
#endif
}
/****************************************************************************
 * cmd_optionValue
 *    returns the option value
 * 
 * input : option character
 *
 * return the option value 
 *
 **************************************************************************/
optionValue_t cmd_optionValue(command_ptr cmdPtr, s08 option)
{
    option_ptr optr = cmdPtr->opt;
    static optionValue_t dummy = {0};
    
    if (optr!=NULL)
    {
        while (optr->key)
        {
            if (optr->key==option)
                return optr->val;
            optr++;
        }
    }
    return dummy; /* error */
}

/****************************************************************************
 * cmd_optionEntered
 *    returns true the option is entered by user
 * 
 * input : option character
 *
 * return the option value 
 *
 **************************************************************************/
u08 cmd_optionEntered(command_ptr cmdPtr, s08 option)
{
    option_ptr optr = cmdPtr->opt;
    
    if (optr!=NULL)
    {
        while (optr->key)
        {
            if (optr->key==option)
                return optr->entered;
            optr++;
        }
    }
    return 0; 
}

/****************************************************************************
 * cmd_prtitle
 *    print out command help
 * 
 * input : command structure 
 *
 * return : ERR_INVALID
 **************************************************************************/
void cmd_prTitle(command_ptr cmdPtr)
{
     int len = 15 - strlen(cmdPtr->cmd);
     
     win_setTextColor(0xc);
     printf("%s",cmdPtr->cmd);
     win_setTextColor(WHITE);
     while (len--) printf(" ");
     printf("%s\n",cmdPtr->title);
     win_setTextColor(GRAY);
}
    
/****************************************************************************/
void cmd_SetIteration(command_ptr pCmd, int itr)
{
    option_ptr optr = pCmd->opt;
    
    printf("setting iteration with value %d\n",itr);
    if (optr!=NULL)
    {
        while (optr->key)
        {
            if (optr->type==OPT_LOOP)
            {
                optr->entered = 1;
                optr->val.valueInt = itr;
                break;
            }
            optr++;
        }
    }
}

/****************************************************************************
 * cmd_help
 *    print out command help
 * 
 * input : command structure 
 *
 * return : ERR_INVALID
 **************************************************************************/
int cmd_help(command_ptr cmdPtr)
{
    option_ptr optr=cmdPtr->opt;
    int count = 0;
    
    cmd_prTitle(cmdPtr);
    printf("\nusage: %s %s\n",cmdPtr->cmd,cmdPtr->syntax);
    if (cmdPtr->help && cmdPtr->help[0])
    {
        win_setTextColor(WHITE);
        printf("\nDescription:\n");
        win_setTextColor(GRAY);
        printf("%s\n\n",cmdPtr->help);
    }
    if (optr!=NULL)
    {
        win_setTextColor(WHITE);
        printf("Option:\n");
        win_setTextColor(GRAY);
        while (optr->key)
        {
            if (++count > 20)
            {
                util_pause(NULL);
                count = 0;
            }
            printf("   -%c",optr->key);
            switch (optr->type) 
            {
                case OPT_U32:
                case OPT_INT:
                    printf("<HEX>     %s ",optr->help);
                    if (optr->defVal.value32!=NO_DEFAULT)
                        printf("(def=%08X)\n",optr->defVal);
                    else 
                        printf("\n");
                    break;
                case OPT_DEC:
                    printf("<DEC>     %s ",optr->help);
                    if (optr->defVal.value32!=NO_DEFAULT)
                        printf("(def=%d)\n",optr->defVal);
                    else 
                        printf("\n");
                    break;
                case OPT_LOOP:
                    printf("<DEC>     %s\n",optr->help);
                    break;
                case OPT_STR:
                    printf("<string>  %s\n",optr->help);
                    break;
                default:
                    printf("          %s\n",optr->help);
            }
            optr++;
        }
    }
    return ERR_INVALID;
} 

/****************************************************************************
 * cmd_disp_helps
 *    print out all commands available
 * 
 * input : command structure 
 *
 * return : ERR_INVALID
 **************************************************************************/
int cmd_disp_helps(command_ptr * Cmds, int grp, int endGrp)
{
    int cnt;
    command_ptr * cmd;
    
    cnt = 0;
    while (grp <= endGrp)
    {
        int titlePrinted = 0;

        cmd = Cmds;
        while (*cmd)
        {
            if ((*cmd)->group==grp)
            {
                if (!titlePrinted)
                {
                    win_setTextColor(HELP_COLOR);
                    printf("\nCommand Group '%s'\n",cmd_grp_str[grp]);
                    cnt+=2;
                    win_setTextColor(GRAY);
                    titlePrinted = 1;
                }
                cmd_prTitle(*cmd);
                if (cnt++>=(win_viewRowSize()-4))
                {
                    if (util_pause(NULL)==KEY_ESC)
                    {
                        printf("\n");
                        return ERR_ABORT;
                    }
                    printf("\n");
                    cnt = 0;
                }
            }    
            cmd++;
            check_abort();
        }
        grp++;
    }
    printf("\n");
    return NO_ERROR;
}

/****************************************************************************
 * cmd_clearOption
 *    clear all option value
 * 
 * input : command structure 
 *
 **************************************************************************/
void cmd_clearOption(command_ptr cmdPtr)
{
    option_ptr optr=cmdPtr->opt;
    
    if (optr!=NULL)
    {
        while (optr->key)
        {
            optr->val.value32 = optr->defVal.value32 == NO_DEFAULT ? 0 : optr->defVal.value32;
            optr->entered = 0;
            optr++;
        }
    }
} 

