/********************************************************************
 * Copyright(c) 2000 Broadcom Corporation, all rights reserved.
 * Contains proprietary and confidential information.
 *
 * This source file is the property of Broadcom Corporation, and
 * may not be copied or distributed in any isomorphic form without 
 * the prior written consent of Broadcom Corporation. 
 *
 *  FileName: T3_COMM.C
 *
 *  Description: This file contains rs232 interface routines 
 *
 ********************************************************************/ 

#include <stdio.h>
#ifndef EFI
#include <conio.h>
#endif

#include "general.h"
#include "t3_gdata.h"
#include "t3_comm.h"

#define COM_DLL  0
#define COM_DLM  1
#define COM_DATA 0
#define COM_IER  1
#define COM_FCR  2
#define COM_IIR  2
#define COM_LCR  3
#define COM_MCR  4
#define COM_LSR  5
#define COM_MSR  6
#define COM_SCR  7

#define COM_SPEED_1200      0x60
#define COM_SPEED_2400      0x30
#define COM_SPEED_4800      0x18
#define COM_SPEED_9600      0x0C
#define COM_SPEED_19200     0x06
#define COM_SPEED_28800     0x04
#define COM_SPEED_38400     0x03
#define COM_SPEED_57600     0x02
#define COM_SPEED_115200    0x01

#define COM_IER_RX_DATA_READY   0x01
#define COM_IER_TX_EMPTY        0x02
#define COM_IER_RX_LINE_STATUS  0x04
#define COM_IER_RX_MODEM_STATUS 0x08

// 3: length 8
// 2: length 7
// 1: length 6
// 0: length 5
#define COM_LCR_WORD_LENGTH_8           0x03
// Stop bit: 1:2 stop bit, 0:1 stop bit
#define COM_LCR_STOP_BIT                0x04
// parity bit 4:5
//    0: none   -- PARITY DISABLED
//    1: ODD
//    2: EVEN
//    3: MARK
//    5: MARK
//    7: SPACE
#define COM_LCR_EVEN_PARITY                 0x18
#define COM_LCR_DIVISOR_LATCH_ACCESS_BIT    0x80

#define COM_MCR_DTR             0x01
#define COM_MCR_RTS             0x02
#define COM_MCR_OUT1            0x04
#define COM_MCR_OUT2            0x08
#define COM_MCR_LOCAL_LOOPBACK  0x10

#define COM_LSR_RX_DATA_READY   0x01
#define COM_LSR_OVERRUN         0x02
#define COM_LSR_PARITY_ERROR    0x04
#define COM_LSR_FRAMING_ERROR   0x08
#define COM_LSR_BREAK_INTERRUPT 0x10
#define COM_LSR_TX_BUFFER_EMPTY 0x20
#define COM_LSR_TX_EMPTY        0x40
#define COM_LSR_RX_FIFO_ERROR   0x80

#define COM_MSR_DELTA_CTS       0x01
#define COM_MSR_DELTA_DSR       0x02
#define COM_MSR_TRAILING_RI     0x04
#define COM_MSR_DELTA_DCD       0x08
#define COM_MSR_CTS             0x10
#define COM_MSR_DSR             0x20
#define COM_MSR_RI              0x40
#define COM_MSR_DCD             0x80

#define COM_TIMEOUT             0x100000

typedef struct {
    u32 base_address;
    int irq;
    int disabled;
} comm_stuct;

static comm_stuct com_global = {0, 0, 0};    

/*************************************************************/
static void configure(int LCR_content)
{
#ifndef EFI
    outp(com_global.base_address+COM_LCR,LCR_content);
#endif    
}

/*************************************************************/
static void commreset()
{
#ifndef EFI
    inp(com_global.base_address+COM_MSR);  // clear MSR
    inp(com_global.base_address+COM_LSR);  // clear LSR
//    set_speed(COM_SPEED_115200); // default speed
//    set_speed(COM_SPEED_1200); // default speed
    configure(COM_LCR_WORD_LENGTH_8); // 8 BIT, no parity, 1 stop
//    configure(COM_LCR_WORD_LENGTH_8|COM_LCR_EVEN_PARITY); // 8 BIT, even parity, 1 stop
    if (com_global.irq)  // hook interrupt service routine to irq
        outp(com_global.base_address+COM_IER,0xff); // enable interrupts
    else
        outp(com_global.base_address+COM_IER,0); // disable interrupts
    inp(com_global.base_address+COM_LSR);      // clear all status
    outp(com_global.base_address+COM_MCR,0xb); // turn on control lines
#endif    
}

/*************************************************************/
static set_speed(u08 new_speed)
{
#ifndef EFI
    outp(com_global.base_address+COM_LCR,
         (unsigned char) inp(com_global.base_address+COM_LCR) | (unsigned char) COM_LCR_DIVISOR_LATCH_ACCESS_BIT);
    outp(com_global.base_address+COM_DLM,0);
    outp(com_global.base_address+COM_DLL,(unsigned char) new_speed);
    outp(com_global.base_address+COM_LCR,
         (unsigned char) inp(com_global.base_address+COM_LCR) & 0x7f); // clear access bit
#endif    
}

/*************************************************************/
static int tx_ready(void)
{
#ifndef EFI
    unsigned char status;
    int timeCnt = COM_TIMEOUT;
   
    if (com_global.base_address)
    {
        if (com_global.disabled)
            return 1;
            
        while (--timeCnt)
        {
            if ((status = inp(com_global.base_address+COM_LSR))==0xff)
                break;
            if (status & (COM_LSR_TX_EMPTY | COM_LSR_TX_BUFFER_EMPTY))
            return 1;
        }
    }    
#endif    
    return 0;
}

/*************************************************************/
/********* public routines ***********************************/
/*************************************************************/
int com_isChar()
{
#ifndef EFI
    if (com_global.base_address)
    {
        unsigned char status = inp(com_global.base_address+COM_LSR);
        if (status == 0xff)
          return 0;
        return (status & COM_LSR_RX_DATA_READY);
    }
#endif    
    return 0;
}

/*************************************************************/
int com_init(u32 port, u32 speed)
{
#ifndef EFI
    u32 base[4] = {0x3f8, 0x2f8, 0x3e8, 0x2e8};
    u08 sp;

    com_global.base_address = 0;
    switch (speed) 
    {
        case 9600:
            sp = COM_SPEED_9600;
            break;
        case 19200:
            sp = COM_SPEED_19200;
            break;
        case 38400:
            sp = COM_SPEED_38400;
            break;
        case 57600:
            sp = COM_SPEED_57600;
            break;
        case 115200:
            sp = COM_SPEED_115200;
            break;
        default:
            eprintf("Invalid Speed\n");
            return 0;
    }    
    
    if (port < 4)
    {
        com_global.base_address = base[port];
        commreset();
        set_speed(sp);
        return 1;
    }
    else
        eprintf("Invalid Port, must be (0..3)\n");
    return 0;
#else
    return 1;    
#endif
}

/*************************************************************/
int com_getch()
{
#ifndef EFI
    int timeCnt = COM_TIMEOUT;
    u08 data;
    
    if (!com_global.base_address)
        return 0;
    while (--timeCnt && !com_isChar());
    if (!timeCnt)
        return 0;
    data = inp(com_global.base_address);
    return (int) data;
#else
    return 0;    
#endif
}

/*************************************************************/
int com_putch(char c)
{
#ifndef EFI
    if (tx_ready() && !com_global.disabled)
    {
        outp(com_global.base_address,c);
        if (c=='\n')
            com_putch('\r');    
        return 1;
    }
    return 0;
#else
    return 1;
#endif    
}

/*************************************************************/
int com_puts(char * str)
{
#ifndef EFI
    if (com_global.base_address && !com_global.disabled)
    {   
        while (*str)
        {
            if (!com_putch(*str++))
                return 0;
        }
        return 1;
    }
    return 0;
#else
    return 1;    
#endif
}

/*************************************************************/
void com_disable(void)
{
    com_global.disabled = 1;
}

/*************************************************************/
void com_enable(void)
{
    com_global.disabled = 0;
}
