/********************************************************************
 * Copyright(c) 2000 Broadcom Corporation, all rights reserved.
 * Contains proprietary and confidential information.
 *
 * This source file is the property of Broadcom Corporation, and
 * may not be copied or distributed in any isomorphic form without 
 * the prior written consent of Broadcom Corporation. 
 *
 *
 *  Name: T 3 _ F L A S H . C
 *
 *  Description: This file contains serial Flash/EEPROM utility routines
 *               for 5703.
 *
 *  AUTHOR: Stephen Keung
 *
 *******************************************************************/

#ifdef NIC_CPU_VIEW
    #include "types.h"
    #include "t3.h"
    #include "t3_misc.h"

#else
    #include <stdlib.h>
    #include "t_io.h"
    #include "t3_util.h"
    #include "general.h"
    #include "t3_gdata.h"
    #include "t3_win.h"
    #define PRINT  printf
    #define PRINT1 printf
    #define PRINT2 printf
    #include "bitbang.h"
#endif
#include "t3_flash.h"

#define FLASH_TYPE (FLASH_CFG1 & FLSH_TYPE_MASK)
#define NON_BUF_FLASH_SECTOR_MASK 0xFFFFF000
#define NON_BUF_FLASH_BYTE_MASK 0xFFF
#define NON_BUF_FLASH_SECTOR_SIZE 0x1000

/* Buffered flash (Atmel: AT45DB011B) specific information */
#define BUFFERED_FLASH_PAGE_POS 9
#define BUFFERED_FLASH_BYTE_ADDR_MASK ((1<<BUFFERED_FLASH_PAGE_POS) - 1)
#define BUFFERED_FLASH_PAGE_SIZE 264
#define BUFFERED_FLASH_PHY_PAGE_SIZE 512


#define FLASH_WAIT 1000000

#define SERIAL_WRITE (u08)0xa0
#define SERIAL_READ  (u08)0xa1

/*****************************************************************/
/************** private procedures *******************************/
/*****************************************************************/

/******************************************************/
static void flshEepSetClock(int value)
{
    u32 data;

    /* enable SCL as output pin */
    t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Addr), 
                 t_readmem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Addr)) | FLSH_BITBANG_MODE_SCL);
    data = t_readmem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Write));
    t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Write), 
                value ? data | FLSH_BITBANG_MODE_SCL : data & ~FLSH_BITBANG_MODE_SCL);
}

/******************************************************/
static void flshEepSetData(int value)
{
    u32 data;

    /* enable SDA as output pin */
    t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Addr), 
                 t_readmem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Addr)) | FLSH_BITBANG_MODE_SDA);
    data = t_readmem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Write));
    t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Write), 
                value ? data | FLSH_BITBANG_MODE_SDA : data & ~FLSH_BITBANG_MODE_SDA);
}

/******************************************************/
static int flshEepGetData(void)
{
    /* change SDA to input pin */
    t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Addr), 
                 t_readmem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Addr)) & ~FLSH_BITBANG_MODE_SDA);
    return t_readmem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Read)) & FLSH_BITBANG_MODE_SDA;
}

/******************************************************/
static void flshFlsSetClock(int value)
{
    u32 data;

    /* enable SCL as output pin */
    t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Addr), 
                 t_readmem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Addr)) | FLSH_BITBANG_MODE_SCK);
    data = t_readmem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Write));
    t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Write), 
                value ? data | FLSH_BITBANG_MODE_SCK : data & ~FLSH_BITBANG_MODE_SCK);
}

/******************************************************/
static void flshFlsSetData(int value)
{
    u32 data;

    /* enable SDA as output pin */
    t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Addr), 
                 t_readmem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Addr)) | FLSH_BITBANG_MODE_SO);
    data = t_readmem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Write));
    t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Write), 
                value ? data | FLSH_BITBANG_MODE_SO : data & ~FLSH_BITBANG_MODE_SO);
}

/******************************************************/
static int flshFlsGetData(void)
{
    /* change SDA to input pin */
    t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Addr), 
                 t_readmem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Addr)) & ~FLSH_BITBANG_MODE_SI);
    return t_readmem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Read)) & FLSH_BITBANG_MODE_SI;
}

/******************************************************/
static void flshFlsSetCS(int value)
{
    u32 data;

    /* enable SDA as output pin */
    t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Addr), 
                 t_readmem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Addr)) | FLSH_BITBANG_MODE_CS_L);
    data = t_readmem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Write));
    t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Write), 
                value ? data | FLSH_BITBANG_MODE_CS_L : data & ~FLSH_BITBANG_MODE_CS_L);
}

/******************************************************/
static int erase_sector(u32 sector, int verbose)
{
    u32 cfg = FLASH_CFG1;
    int j, result;

    if (lockFlash())
    {
        if (cfg & FLSH_CFG_BITBANG_MODE_ENABLE)
            result = sstBitBangSectErase(sector, flshFlsSetClock, flshFlsSetData, flshFlsGetData, flshFlsSetCS);
        else
        {
            /* Erase the sector */
            t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Addr),sector);
            t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Command),
                     FLSH_CMD_DOIT | FLSH_CMD_LAST | FLSH_CMD_FIRST | 
                     FLSH_CMD_DONE | FLSH_CMD_WR | FLSH_CMD_ERASE);

            result = HAS_ERROR;
            for (j = 0; j < (FLASH_WAIT);j++)
                if (t_readmem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Command)) & FLSH_CMD_DONE)
                {
                    result = NO_ERROR;
                    break;
                }    
        }
        if (result!=NO_ERROR)
            eprintf("Flash sector erase timeout\n");
        releaseFlash();
    }
    return result;
}

/******************************************************/
static int flash_bitbang_write_word(u32 addr, u32 data, U32 first_last, int type)
{
    u08 dev;
    int result;
    
    if (lockFlash())
    {
        switch (type) {
        case FLSH_TYPE_NO_BUFFER_FLASH:
            result = sstBitBangWriteWord(addr, data, flshFlsSetClock, flshFlsSetData, flshFlsGetData, flshFlsSetCS);
            break;
            
        case FLSH_TYPE_BUFFERED_FLASH:
            result = atmlBitBangWriteWord(addr, data, first_last, flshFlsSetClock, flshFlsSetData, flshFlsGetData, flshFlsSetCS);
            break;
            
        case FLSH_TYPE_JUST_SEEPROM:
            dev = (u08) (addr >> 16);
            result = eepBitBangWriteWord(dev, (u16) addr, data, flshEepSetClock, flshEepSetData, flshEepGetData);
            break;
            
        default:
            result = HAS_ERROR;
            break;
        }
        releaseFlash();
    }
    return result;
}    

/******************************************************/
static void setupSeepromDeviceID(u32 addr)
{
    u32 dev;
    u32 data;
    
    /* initialize dev id in read/write command */
    dev = (addr >> 15) & 0x6;      /* device id * 2 */
    data = t_readmem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Config3)) & ~FLSH_SEE_CMD_MASK;
    data |= ((dev | SERIAL_READ) << FLSH_SEE_READ_CMD_POS) | ((dev | SERIAL_WRITE) << FLSH_SEE_WRITE_CMD_POS);
}

/******************************************************/
static int flash_auto_write_word(u32 addr, u32 data, u32 first_last, int flash)
{            
    int i;

    if (!flash)
        setupSeepromDeviceID(addr);
    
    if (lockFlash())
    {
        /* Fill the data to be written */
        t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Write),data);

        /* Fill the address */
        if (first_last & FLSH_CMD_FIRST)
            t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Addr),addr);

        /* Issue the doit command */
        t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Command),
               first_last | FLSH_CMD_DONE | FLSH_CMD_DOIT | FLSH_CMD_WR);

        for (i = 0; i < FLASH_WAIT;i++)
            if (t_readmem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Command)) & FLSH_CMD_DONE)
            {
                releaseFlash();
                return NO_ERROR;
            }    
   
        releaseFlash();
        eprintf("Flash Auto Write timeout error\n");
    }    
    return HAS_ERROR;
}        

/******************************************************/
static int flash_write_word(u32 flash_addr, u32 wr_data, u32 first_last)
{
    u32 cfg = FLASH_CFG1;

    return (cfg & FLSH_CFG_BITBANG_MODE_ENABLE) ?
        flash_bitbang_write_word(flash_addr, wr_data, first_last, cfg & FLSH_TYPE_MASK) :
        flash_auto_write_word(flash_addr, wr_data, first_last, cfg & FLSH_CFG_FLASH_MODE_ENABLE);
}

/******************************************************/
static int flash_read_swappwed_word(u32 ad, u32 * buf)
{
    int rtv = flash_read_word(ad,buf);
#ifndef NIC_CPU_VIEW
    *buf = BYTE_SWAP(*buf);
#endif
    return rtv;
}

/******************************************************/
static void pr_num(int verbose, u32 ad, int col)
{
#ifndef NIC_CPU_VIEW
    if (verbose)
    {
        win_gotoCol(col);
        PRINT1("%08X",ad);
    }
#endif    
}

/******************************************************/
static void pr_msg(int verbose, char * st, int col)
{
#ifndef NIC_CPU_VIEW
    if (verbose)
    {
        win_gotoCol(col);
        PRINT(st);
    }
#endif
}

static u32 sectorBuf[NON_BUF_FLASH_SECTOR_SIZE/4];

/******************************************************/
/* len unit is in byte                                */
/* data will swapped to big endian                    */
/******************************************************/
static int programNoBufferFlash(u32 beginAddr, u32 num_word, u32 * buff, int verbose)
{
    u32 ad;
    u32 endAddr = beginAddr + num_word*4;
    u32 sector, end_sector;
    u32 erased;
    int begIdx,idx,i;
    u32 dataIdx = 0;
#ifndef NIC_CPU_VIEW    
    int row;
#endif    
    int col2,col;
    
    /* Fill it sector by sector */
    sector = beginAddr & NON_BUF_FLASH_SECTOR_MASK;
    end_sector = (endAddr-1) & NON_BUF_FLASH_SECTOR_MASK;
    
    for (; sector <= end_sector; sector+=NON_BUF_FLASH_SECTOR_SIZE) 
    {
        if (verbose)
        {
            PRINT1("\n  Sector %02X: ", beginAddr >> 12);
            win_wherexy(&row,&col);
            printf("Verifying:");
            win_wherexy(&row,&col2);
        }
        /* if program area is all 0xffffffff (erased), we just need to program 
           that area */
        ad = sector < beginAddr ? beginAddr : sector;
        erased = 1;
        begIdx = (ad-sector)/4;
        
        for (idx=begIdx; ad < (sector + NON_BUF_FLASH_SECTOR_SIZE) && (ad < endAddr); ad+=4, idx++, dataIdx++)
        {
            if (erased)
            {
                pr_num(verbose, ad, col2);
                if (flash_read_swappwed_word(ad, sectorBuf + idx)!=NO_ERROR)
                    return HAS_ERROR;
                if (sectorBuf[idx] != 0xffffffff)
                    erased = 0;
            }
            sectorBuf[idx] = BYTE_SWAP(buff[dataIdx]);
        }
        
        if (!erased) /* need to erase sector and merge with exisiting data (if any) */
        {
            pr_msg(verbose,"Reading..:",col);
            
            /* save data before erase */
            for (i=0,ad=sector; i<begIdx; i++,ad+=4)
            {
                pr_num(verbose, ad, col2);
                if (flash_read_swappwed_word(ad,sectorBuf+i)!=NO_ERROR)
                    return HAS_ERROR;
            }
            for (ad=sector+idx*4; idx<(NON_BUF_FLASH_SECTOR_SIZE/4); idx++,ad+=4)
            {
                pr_num(verbose, ad, col2);
                if (flash_read_swappwed_word(ad,sectorBuf+idx)!=NO_ERROR)
                    return HAS_ERROR;
            }
            pr_msg(verbose,"Erasing..:",col);
            if (erase_sector(sector,verbose)!=NO_ERROR)
                return HAS_ERROR;
                
            begIdx = 0;
            idx = NON_BUF_FLASH_SECTOR_SIZE/4;
        }

        pr_msg(verbose,"Writing..:",col);
        for (ad=sector+begIdx*4;begIdx < idx;begIdx++,ad+=4)
        {
            pr_num(verbose, ad, col2);
            if (flash_write_word(ad,sectorBuf[begIdx], FLSH_CMD_FIRST | FLSH_CMD_LAST)!=NO_ERROR)
                return HAS_ERROR;
        }
        pr_msg(verbose,"Done              ",col);
    }
    pr_msg(verbose,"\n",col);
    return NO_ERROR;
}

/******************************************************/
static int programSEE(u32 address, u32 num_word, u32 * buff,int verbose, int type)
{
    u32 i, prog_addr;
    u32 start_addr = address;
    u32 end_addr = start_addr + (num_word - 1) * 4;
    u32 control;
#ifndef NIC_CPU_VIEW
    int row, col;
#endif    

    if (verbose==2)
    {
        printf("Programming... ");    
    }
    win_wherexy(&row,&col);
    for (i=0 ; i<num_word; i++, address+=4)
    {
        if (type==FLSH_TYPE_BUFFERED_FLASH)
        {
            prog_addr = flash_LogicalToPhysicalAddress(address);
            control = !(prog_addr & BUFFERED_FLASH_BYTE_ADDR_MASK) || 
                       (address == start_addr) ? FLSH_CMD_FIRST : 0;
            if ((prog_addr & BUFFERED_FLASH_BYTE_ADDR_MASK)>=(BUFFERED_FLASH_PAGE_SIZE-4) || (address==end_addr))
                control |= FLSH_CMD_LAST;
        }
        else
        {
            prog_addr = address;
            control = FLSH_CMD_FIRST | FLSH_CMD_LAST;
        }
        
        if (flash_write_word(prog_addr, BYTE_SWAP(buff[i]), control)!=NO_ERROR)
            return HAS_ERROR;
        if (verbose)
        {    
            win_gotoCol(col);
            PRINT1("%5u bytes", i*4);
        }
    }
    if (verbose)
    {
        win_gotoCol(col);
        PRINT1("%5u bytes\n", i*4);
    }    
    return NO_ERROR;
}

/******************************************************/
static int readFlashBitBangMode(u32 addr, u32* data, int type)
{
    u08 dev;
    
    switch (type) {
    case FLSH_TYPE_NO_BUFFER_FLASH:
        return sstBitBangReadWord(addr, data, 1, flshFlsSetClock, flshFlsSetData, flshFlsGetData, flshFlsSetCS);
    case FLSH_TYPE_BUFFERED_FLASH:
        return atmlBitBangReadWord(addr, data, flshFlsSetClock, flshFlsSetData, flshFlsGetData, flshFlsSetCS);
    case FLSH_TYPE_JUST_SEEPROM:
        dev = (u08) (addr >> 16);
        return eepBitBangReadWord(dev, (u16) addr, data, flshEepSetClock, flshEepSetData, flshEepGetData);
    default:
        return HAS_ERROR;
    }
}    

/******************************************************/
static int readFlashPassThouMode(u32 flash_addr, u32* data)
{
    eprintf("*** Error ***, procedure not ready\n");
    *data = 0xffffffff;
    return HAS_ERROR;
}    

/******************************************************/
static int readFlashAutoMode(u32 addr, u32* data, int flash)
{ 
    int i;
               
    if (!flash)
        setupSeepromDeviceID(addr);
    
    t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Addr), addr & FLSH_ADDRESS_MASK);

    /* Issue the doit command, we also need to write FLSH_CMD_DONE to clear pervious status */
    t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Command),
           FLSH_CMD_DOIT | FLSH_CMD_FIRST | FLSH_CMD_LAST | FLSH_CMD_RD | FLSH_CMD_DONE);
    
    t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Command),
           FLSH_CMD_DOIT | FLSH_CMD_FIRST | FLSH_CMD_LAST | FLSH_CMD_RD | FLSH_CMD_DONE);

    /* Wait for the done bit */
    for (i = 0; (i < FLASH_WAIT) && !util_user_abort(); i++)
    {
        if (t_readmem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Command)) & FLSH_CMD_DONE)
        {
            *data = t_readmem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Read));
            t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.Command),FLSH_CMD_DONE);
#ifndef NIC_CPU_VIEW
                    *data = BYTE_SWAP(*data);
#endif
            return NO_ERROR;
        }
    }
    PRINT("Flash read timout error\n");
    return HAS_ERROR;    
}
        
/*****************************************************************/
/************** public procedures ********************************/
/*****************************************************************/

#ifndef NIC_CPU_VIEW
/*********************************************************************************/
/* buffered flash has 512 byte pages contains only 264 bytes                     */
/* Therefore, there is 'hold' in each pages as physical address                  */
/* this routine converts from logical address to buffered flash physical address */
/*********************************************************************************/
u32 flash_LogicalToPhysicalAddress(u32 address)
{
    u32 ad = address / BUFFERED_FLASH_PAGE_SIZE;
    return (ad << BUFFERED_FLASH_PAGE_POS) + (address % BUFFERED_FLASH_PAGE_SIZE);
}

/*********************************************************************************/
/* buffered flash has 512 byte pages contains only 264 bytes                     */
/* Therefore, there is 'hold' in each pages as physical address                  */
/* this routine converts from logical address to buffered flash physical address */
/*********************************************************************************/
u32 flash_PhysicalToLogicalAddress(u32 address)
{
    return (address >> BUFFERED_FLASH_PAGE_POS) * BUFFERED_FLASH_PAGE_SIZE + 
           (address & BUFFERED_FLASH_BYTE_ADDR_MASK);
}

#else

/*********************************************/
u32 flash_LogicalToPhysicalAddress(u32 address)
{
/*
    since internal CPU do not support mutliple or devide instruction, so we have to 
    do it in hard way...
*/
    u32 ad = 0;
    while (address >= BUFFERED_FLASH_PAGE_SIZE)
    {
        ad += BUFFERED_FLASH_PHY_PAGE_SIZE;
        address -= BUFFERED_FLASH_PAGE_SIZE;
    }
    return ad + address;
/*    
    u32 ad = address / BUFFERED_FLASH_PAGE_SIZE;
    return (ad << BUFFERED_FLASH_PAGE_POS) + (address % BUFFERED_FLASH_PAGE_SIZE);
*/    
}
#endif

/******************************************************/

int flash_program(u32 addr, u32 num_word, u32 * buff, int verbose)
{
    u32 type = FLASH_TYPE;
    
    switch (type) {
    case FLSH_TYPE_NO_BUFFER_FLASH:
        return programNoBufferFlash(addr, num_word, buff, verbose);
    case FLSH_TYPE_BUFFERED_FLASH:
    case FLSH_TYPE_JUST_SEEPROM:
        return programSEE(addr, num_word, buff, verbose, type);
    default:
        return HAS_ERROR;
    }
}

/******************************************************/
int flash_read_word(u32 flash_addr, u32* data)
{
    int result = HAS_ERROR;
    u32 cfg = FLASH_CFG1;

    if (lockFlash())
    {
        if ((cfg & FLSH_TYPE_MASK) == FLSH_TYPE_BUFFERED_FLASH) 
            flash_addr = flash_LogicalToPhysicalAddress(flash_addr);
        
        if (cfg & FLSH_CFG_BITBANG_MODE_ENABLE) /* bit bang mode */
            result = readFlashBitBangMode(flash_addr, data, cfg & FLSH_TYPE_MASK);
        else if (cfg & FLSH_CFG_PASS_MODE_ENABLE)
            result = readFlashPassThouMode(flash_addr, data);
        else
            result = readFlashAutoMode(flash_addr, data, cfg & FLSH_CFG_FLASH_MODE_ENABLE);
        releaseFlash();
    }    
    return result;
}

#ifndef NIC_CPU_VIEW
/******************************************************/
/*
u32 flash_read(u32 addr)
{
    u32 data;
    int failed = flash_read_word(addr,&data);

    / * First of all, convert data to little-endian first * /
    data = BYTE_SWAP(data);
    if (failed != NO_ERROR)
    {
        PRINT1("Error reading from flash rtc = %d\n",failed);
        return 0xffffffff;
    }
    return data;
} */

/******************************************************/
void flash_fill_blk(u32 address, u32 end,u32 wr_data, int verbose)
{
    int i, len = (end-address+4)/4;
    u32 * buffer;
    
    buffer = (u32 *) malloc(len*4);
    for (i=0;i<len;i++)
        buffer[i] = BYTE_SWAP(wr_data);
    flash_program(address, len, buffer, verbose);
    free (buffer);
}

/******************************************************/
/*
int flash_read_blk(u32 address,u32 *buffer,u32 num_words, int verbose)
{
    u32 data;

    while (num_words--)
    {
        if (verbose)
        {
        }
        if (flash_read_word(address, &data) != NO_ERROR)
            return HAS_ERROR;
        *buffer++ = data;
        address += 4;
    }
    return NO_ERROR;
}
*/
/******************************************************/
int flash_isSupported()
{
    return util_is5703_or_later();
}
#else
#define flash_isSupported() 1
#endif

/******************************************************/
/*
int flash_nonBufferedFlash()
{
    return !flash_oldMethod() && (FLASH_TYPE == FLSH_TYPE_NO_BUFFER_FLASH);
}
*/
/******************************************************/
int flash_BufferedFlash()
{
    return !flash_oldMethod() && (FLASH_TYPE == FLSH_TYPE_BUFFERED_FLASH);
}

#ifndef NIC_CPU_VIEW
/******************************************************/
int flash_showMode(void)
{
    u32 cfg_info;
    
    cfg_info = FLASH_CFG1;
    PRINT1("Current mode: New FLASH %s", cfg_info & FLSH_CFG_BITBANG_MODE_ENABLE ?  "bitbang" : "auto");
    if (!(cfg_info & FLSH_CFG_BITBANG_MODE_ENABLE))
        PRINT1("; %spassthru\n", cfg_info & FLSH_CFG_PASS_MODE_ENABLE ?  "" : "no ");
    else
        PRINT("\n");    
    PRINT2("device: %s",cfg_info & FLSH_CFG_FLASH_MODE_ENABLE ?  "Flash" : "EEPROM");
    if (cfg_info & FLSH_CFG_FLASH_MODE_ENABLE)
        PRINT(" with %sbuffer\n", cfg_info & FLSH_CFG_BUFFER_MODE_ENABLE ?  "" : "no ");
    else
        PRINT("\n");
    return NO_ERROR;
}
#endif

/******************************************************/
int flash_oldMethod(void)
{
    return !flash_isSupported() || (FLASH_CFG1 & FLSH_CFG_COMPAT_BYPASS);
}

/******************************************************/
int lockFlash(void)
{
    u32 timeCnt = FLASH_WAIT/4;

    if (flash_isSupported())
    {
        while (--timeCnt && (t_readmem32(OFFSETOF(T3_STD_MEM_MAP,Flash.SwArbit)) & 
                (FLSH_SW_ARBIT_ARB0 | FLSH_SW_ARBIT_ARB1 | FLSH_SW_ARBIT_ARB2 | FLSH_SW_ARBIT_ARB3)))
            if (util_user_abort())
                return 0;
        if (timeCnt)
        {
            t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.SwArbit), FLSH_SW_ARBIT_REQ_SET0);
            for (timeCnt=0; timeCnt< FLASH_WAIT/4; timeCnt++)
                if (t_readmem32(OFFSETOF(T3_STD_MEM_MAP,Flash.SwArbit)) & FLSH_SW_ARBIT_ARB0)
                    return 1;
        }            
        eprintf("Cannot gain access to flash\n");
        return 0;
    }    
    return 1;
}
    
/******************************************************/
void releaseFlash(void)
{
    if (flash_isSupported())
        t_writemem32(OFFSETOF(T3_STD_MEM_MAP,Flash.SwArbit), FLSH_SW_ARBIT_REQ_CLR0);
}




